/* setfsb.c - FSB modification tool
 * Copyright (C) 2003 Georg Lukas <georg@boerde.de>
 * (This Program is under the GNU General Public License)
 *
 * WARNING!!! DO NOT USE THIS PROGRAM!!!
 *
 * If you use it despite the warning, it might ruin your CPU, motherboard or
 * your sex life!
 *
 * Ok, so what does it do? It changes the FSB and PCI clocks on motherboards
 * equipped with the W83194R-39A PLL. This allows you to overclock and crash
 * your machine. Pretty cool, eh? ;)
 *
 * You need a working IC/SMBus (see the lm_sensors package for drivers) to be
 * able to access the W83194R-39A, this program uses the i2c-dev interface for
 * communication with it.
 *
 * I PROVIDE NO WARRANTY WHATSOEVER. IF YOU FSCK UP YOUR MAINBOARD, IT IS ALL
 * YOUR FAULT. YOU'VE BEEN WARNED!
 */

#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
//#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <linux/limits.h>

#include "i2c-workaround.h"
#include "config.h"
#include "pll_desc.h"
#include "pll_list.h"

int verbose = 0;

#ifdef NEED_FUNC
static inline int32_t i2c_smbus_access(int file, char read_write, uint8_t command,
                                     int size, union i2c_smbus_data *data)
{
        struct i2c_smbus_ioctl_data args;

        args.read_write = read_write;
        args.command = command;
        args.size = size;
        args.data = data;
        return ioctl(file,I2C_SMBUS,&args);
}
#endif


/* returns a pointer to the PLL with given name */
pll *find_chip(char *name) {
	struct fsb_mapping *m;
	pll **chip;

	chip = pll_list;
	while (*chip) {
		if (strcmp((*chip)->name, name) == 0)
			return *chip;
		chip++;
	}
	return NULL;
}


/* prints a list of the PLLs supported by fsbset to stdout */
void list_chips() {
	struct fsb_mapping *m;
	pll **chip;
	pll_board *b;

	printf("setfsb supports the following PLL chips:\n");
	chip = pll_list;
	while (*chip) {
		printf("	%s", (*chip)->name);
		if ((*chip)->comment)
			printf(" (%s)", (*chip)->comment);
		printf("\n");
		b = (*chip)->boards;
		while (*b) {
			printf("		%s\n", (*b));
			b++;
		}
		chip++;
	}
}

/* prints a list of the FSBs supported by chip to stdout */
void list_fsbs(pll *chip) {
	pll_fsb_mapping *m;
	pll_feature *f;

	printf("%s supported clocks:\n", chip->name);
	printf("	FSB PCI MEM\n");
	m = chip->fsbs;
	while ((m!=NULL) && (m->fsb != 0)) {
		if (m->fsb != FSB_DEF)
			printf("	%3i %3i %3i\n", m->fsb, m->pci, m->mem);
		m++;
	}
	printf("%s supported features:\n", chip->name);
	f = chip->features;
	while ((f!=NULL) && (f->name != NULL)) {
		printf("	%s\n", f->name);
		f++;
	}
}



/* reads out the PLL registers via SMBus and returns them in buf */
int pll_read(int f, pll *chip, union i2c_smbus_data *buf) {
	int res;

	if (ioctl(f,I2C_SLAVE,chip->i2c_addr) < 0) {
		perror("ioctl(I2C_SLAVE)");
		return -1;
	}
	memset(buf->block, 0, sizeof(*buf));
	res = i2c_smbus_access(f, I2C_SMBUS_READ, 0,
			I2C_SMBUS_BLOCK_DATA, buf);
	if (res != 0) {
		perror("pll_read(): i2c_smbus_access()");
		return -2;
	}
}

/* DANGEROUS! writes PLL registers back to the PLL via SMBus */
int pll_write(int f, pll *chip, union i2c_smbus_data *buf) {
	int res;

	if (ioctl(f,I2C_SLAVE,chip->i2c_addr) < 0) {
		perror("ioctl(I2C_SLAVE)");
		return -1;
	}
	res = i2c_smbus_access(f, I2C_SMBUS_WRITE, 0,
			I2C_SMBUS_BLOCK_DATA, buf);
	if (res != 0) {
		perror("pll_write(): i2c_smbus_access()");
		return -2;
	}
}



/* dumps the registers content of the PLL SMBus data to stdout */
int pll_dump(int f, pll *chip, union i2c_smbus_data *buf) {
	int reg;

	if (chip != NULL)
		printf("Register dump of %s (i2c: 0x%02x):\n	",
				chip->name, chip->i2c_addr);
	for (reg=0; reg < buf->block[0]; reg++) {
		printf("%i=0x%02x ", reg, buf->block[1+reg]);
	}
	printf("\n");
}


/* looks if a PLL cmd matches a given SMBus dump
 * returns:
 *     -1  when not matching
 *      0  when matching
 */
int pll_cmd_match(union i2c_smbus_data *buf, pll_cmd c) {
	if (buf->block[0] <= c.regpos) return -1;
	if (verbose > 1) 
		printf("comparing 0x%02x & ~0x%02x = 0x%02x ?= 0x%02x\n",
			buf->block[1+c.regpos], c.andmask, 
			(buf->block[1+c.regpos] & ~c.andmask),
			c.ormask);
	if ((buf->block[1+c.regpos] & ~c.andmask) == c.ormask) return 0;
	return -1;
}

void pll_cmd_apply(union i2c_smbus_data *buf, pll_cmd c) {
	if (buf->block[0] <= c.regpos) {
		fprintf(stderr, "pll_cmd_apply(): internal error - buffer overflow!\n");
		exit(42);
	}
	buf->block[1+c.regpos] = (buf->block[1+c.regpos] & c.andmask)
					| c.ormask;
}

/* changes the FSB in the SMBus dump to the given value */
int pll_setfsb(pll *chip, union i2c_smbus_data *buf, int fsb) {
	pll_fsb_mapping *m;
	pll_feature *f;

	m = chip->fsbs;
	while ((m!=NULL) && (m->fsb != 0)) {
		if (fsb == m->fsb) {
			pll_cmd_apply(buf, m->c);
			return 0;
		}
		m++;
	}
	return -1;
}


/* prints a user readable dump of the i2c registers to stdout */
int pll_parse(pll *chip, union i2c_smbus_data *buf) {
	pll_fsb_mapping *m;
	pll_feature *f;

	m = chip->fsbs;
	while ((m!=NULL) && (m->fsb != 0)) {
		if (pll_cmd_match(buf, m->c) == 0) {
			if (m->fsb == FSB_DEF)
				printf("defaults (aka unknown)\n");
			else
				printf("FSB=%-3i PCI=%-3i MEM=%-3i\n",
					m->fsb, m->pci, m->mem);
			return;
		}
		m++;
	}
	printf("no match found (PLL description ok? PLL working?)\n");
	/* BETA!
	printf("\nenabled features:\n", chip->name);
	f = chip->features;
	while ((f!=NULL) && (f->name != NULL)) {
		if (pll_cmd_match(buf, f->on))
			printf("	%s\n", f->name);
		f++;
	}
	*/
}

/* prints out a cheap help message */
void usage(char *progname, pll *chip) {
	struct fsb_mapping *m;

	printf("Syntax: %s [<i2c-device>] <PLL> [<FSB>|default|query]\n\n",
			progname);
	if (chip == NULL)
		list_chips();
	else
		list_fsbs(chip);
	exit(1);
}

int main(int argc, char *argv[])
{
	unsigned char filename[PATH_MAX] = I2C_DEV,
		*strerr;
	struct stat fs;
	pll *chip = NULL, *c;
	int do_query = 0, do_list = 0, do_set = 0, do_help = 0;
	int i, fsb, v, res;
	int f;
	union i2c_smbus_data buf;

	if (verbose >= 0) {
		printf(VERSION "\n\n");
	}
	for (i=1; i<argc; i++) {
		/* checking the argument for being a number or a PLL name or
		   a device filename */
		v = strtol(argv[i], &strerr, 0);
		if ((strerr == NULL) || (*strerr == '\0')) {
			fsb = v;
			do_set = 1;
			printf("Using FSB=%i\n", fsb);
		} else if (strcmp(argv[i], "default") == 0) {
			fsb = FSB_DEF;
			do_set = 1;
			printf("Resetting defaults...\n");
		} else if (strcmp(argv[i], "query") == 0) {
			do_query = 1;
			printf("Querying current settings...\n");
		} else if (strcmp(argv[i], "list") == 0) {
			do_list = 1;
			printf("Listing PLL features...\n");
		} else if ((c = find_chip(argv[i])) != NULL) {
			chip = c;
			printf("Using %s PLL\n", chip->name);
		} else if ((stat(argv[i], &fs) == 0) && (S_ISCHR(fs.st_mode))) {
			strncpy(filename, argv[i], PATH_MAX);
			printf("Using device %s\n", filename);
		} else {
			do_help = 1;
			printf("Unknown argument: %s\n", argv[i]);
		}
	}

	if (do_help || do_list || (chip == NULL)) usage(argv[0], chip);

	if ((f = open(filename,O_RDWR)) < 0) {
		fprintf(stderr, "setfsb: Can't open %s: %s\n", filename,
				strerror(errno));
		exit(2);
	}
	
	if (do_query) {
		pll_read(f, chip, &buf);
		pll_dump(f, chip, &buf);
		printf("\nCurrent setting:\t");
		pll_parse(chip, &buf);
	}

	if (do_set) {
		// read out the values from PLL
		pll_read(f, chip, &buf);
		printf("\nOld setting:\t");
		pll_parse(chip, &buf);

		// change FSB to the new value, write back
		pll_setfsb(chip, &buf, fsb);
		pll_write(f, chip, &buf);

		// compare results
		pll_read(f, chip, &buf);
		printf("New setting:\t");
		pll_parse(chip, &buf);
	}
	
	close (f);
}



